`timescale 1ns / 1ps
`include "define.v"
/* Module copied and adapted from HTR Altera project.
 Changes from HTR Altera:
 - changed syntax into Verilog2001 style.
 - loc_cs[3:0] turned into 1-bit LocCS.
 - vme_a[14:0] turned into LocAddr[6:0]
 - vme_data_in[15:0] turned into LocData[7:0]
 - eliminate signals and logic for the second Flash device.
  
 This module writes/reads to/from flash.
*/
module FlashHandler(
input clock, resetn, enable, LocCS,	
input[7:0] LocData,	// maybe we can do 2 bytes at a time, maybe not
input[7:0] FlashDataRead,	// only read 1 byte at a time
input vme_write, X_CONFIG_CS1, 	
input[6:0] LocAddr,	// this is used for commands (see above)

output reg ack, flash_rw,	inprog,			
output FLASH_1_CE, FLASH_1_OE, FLASH_1_WE,
output reg [7:0] FlashDataWrite,	// this data goes out to the flash
output reg [23:0] vme_data_out,		// goes to functionsel
output reg [20:0] flash_addr,	
output reg [5:0] state          );
//
// commands
//
wire reset_address = LocAddr[0];
wire write_flash = LocAddr[1];
wire read_flash = LocAddr[2];
wire reset_flash = LocAddr[3];
wire erase_flash = LocAddr[4];
wire increment_address = LocAddr[5];
wire read_internal_address = LocAddr[6];
//
// flash control line outputs
//
wire flash_1 = enable && LocCS;
reg rce, roe, rwe;
wire ce1 = enable & flash_1 ? rce : 1;
assign FLASH_1_CE = ce1 & X_CONFIG_CS1;
assign FLASH_1_WE = enable & flash_1 ? rwe : 1;
wire oe1 = enable & flash_1 ? roe : 1;
assign FLASH_1_OE = oe1 & X_CONFIG_CS1;

// this flash is a "120ns" guy - so the best thing would be to use a very
// slow clock, then it's pretty safe.  
//
reg[5:0] nstate;
parameter[5:0] WAIT=0, ENABLE=1, 
// read value
	READ_SET_ADD=2, READ_SET_CE=3, READ_SET_OE=4, READ_LATCH=5,
// write (program) value
	WRITE_SET_ADD_CE1=6, WRITE_SET_WE_LOW1=7, 
	WRITE_SET_WE_HIGH1=8, WRITE_SET_CE_HIGH1=9, 
	WRITE_SET_ADD_CE2=10, WRITE_SET_WE_LOW2=11, 
	WRITE_SET_WE_HIGH2=12, WRITE_SET_CE_HIGH2=13,
	WRITE_SET_ADD_CE3=14, WRITE_SET_WE_LOW3=15, 
	WRITE_SET_WE_HIGH3=16, WRITE_SET_CE_HIGH3=17,
	WRITE_SET_ADD_CE4=18, WRITE_SET_WE_LOW4=19, 
	WRITE_SET_WE_HIGH4=20, WRITE_SET_CE_HIGH4=21, 
// write (program) erase
	WRITE_SET_ADD_CE_E1=22, WRITE_SET_WE_LOW_E1=23, 
	WRITE_SET_WE_HIGH_E1=24, WRITE_SET_CE_HIGH_E1=25, 
	WRITE_SET_ADD_CE_E2=26, WRITE_SET_WE_LOW_E2=27, 
	WRITE_SET_WE_HIGH_E2=28, WRITE_SET_CE_HIGH_E2=29, 
	WRITE_SET_ADD_CE_E3=30, WRITE_SET_WE_LOW_E3=31, 
	WRITE_SET_WE_HIGH_E3=32, WRITE_SET_CE_HIGH_E3=33, 
	WRITE_SET_ADD_CE_E4=34, WRITE_SET_WE_LOW_E4=35, 
	WRITE_SET_WE_HIGH_E4=36, WRITE_SET_CE_HIGH_E4=37, 
	WRITE_SET_ADD_CE_E5=38, WRITE_SET_WE_LOW_E5=39, 
	WRITE_SET_WE_HIGH_E5=40, WRITE_SET_CE_HIGH_E5=41, 
	WRITE_SET_ADD_CE_E6=42, WRITE_SET_WE_LOW_E6=43, 
	WRITE_SET_WE_HIGH_E6=44, WRITE_SET_CE_HIGH_E6=45, 
// write (program) reset
	WRITE_SET_ADD_CE_R=46, WRITE_SET_WE_LOW_R=47, 
	WRITE_SET_WE_HIGH_R=48, WRITE_SET_CE_HIGH_R=49, 
// misc
	READ_INTERNAL_ADDRESS=50, INCREMENT=51,
	RESET=52, DONE=53;
always @ (posedge clock or negedge resetn)
	if (~resetn) state <= WAIT;
	else state <= #1 nstate;
always @ (state or reset_address or vme_write or read_flash or
	write_flash or reset_flash or erase_flash or enable)
	case (state)
		WAIT:
			if (enable) nstate = ENABLE;
			else nstate = WAIT;
		ENABLE: begin	
			if (reset_address) nstate = RESET;
			else if (vme_write & read_flash ) nstate = READ_SET_ADD;
			else if (~vme_write & write_flash ) nstate = WRITE_SET_ADD_CE1;
			else if (~vme_write & reset_flash) nstate = WRITE_SET_ADD_CE_R;
			else if (~vme_write & erase_flash) nstate = WRITE_SET_ADD_CE_E1;
			else if (vme_write & read_internal_address) nstate = READ_INTERNAL_ADDRESS;
			else if (~vme_write & increment_address) nstate = INCREMENT;
			else nstate = DONE;
			end
		RESET: nstate = DONE;
		READ_INTERNAL_ADDRESS: nstate = DONE;
// increment
		INCREMENT: nstate = DONE;
// read
		READ_SET_ADD:	nstate = READ_SET_CE;
		READ_SET_CE:	nstate = READ_SET_OE; 
		READ_SET_OE:	nstate = READ_LATCH;
		READ_LATCH:		nstate = DONE;
// program 1st
		WRITE_SET_ADD_CE1:	nstate = WRITE_SET_WE_LOW1;
		WRITE_SET_WE_LOW1:	nstate = WRITE_SET_WE_HIGH1;
		WRITE_SET_WE_HIGH1:	nstate = WRITE_SET_CE_HIGH1;
		WRITE_SET_CE_HIGH1:	nstate = WRITE_SET_ADD_CE2;
// program 2nd
		WRITE_SET_ADD_CE2:	nstate = WRITE_SET_WE_LOW2;
		WRITE_SET_WE_LOW2:	nstate = WRITE_SET_WE_HIGH2;
		WRITE_SET_WE_HIGH2:	nstate = WRITE_SET_CE_HIGH2;
		WRITE_SET_CE_HIGH2:	nstate = WRITE_SET_ADD_CE3;
// program 3rd
		WRITE_SET_ADD_CE3:	nstate = WRITE_SET_WE_LOW3;
		WRITE_SET_WE_LOW3:	nstate = WRITE_SET_WE_HIGH3;
		WRITE_SET_WE_HIGH3:	nstate = WRITE_SET_CE_HIGH3;
		WRITE_SET_CE_HIGH3:	nstate = WRITE_SET_ADD_CE4;
// program 4th
		WRITE_SET_ADD_CE4:	nstate = WRITE_SET_WE_LOW4;
		WRITE_SET_WE_LOW4:	nstate = WRITE_SET_WE_HIGH4;
		WRITE_SET_WE_HIGH4:	nstate = WRITE_SET_CE_HIGH4;
		WRITE_SET_CE_HIGH4:	nstate = DONE;
// erase 1st
		WRITE_SET_ADD_CE_E1:	nstate = WRITE_SET_WE_LOW_E1;
		WRITE_SET_WE_LOW_E1:	nstate = WRITE_SET_WE_HIGH_E1;
		WRITE_SET_WE_HIGH_E1:	nstate = WRITE_SET_CE_HIGH_E1;
		WRITE_SET_CE_HIGH_E1:	nstate = WRITE_SET_ADD_CE_E2;
// erase 2nd
		WRITE_SET_ADD_CE_E2:	nstate = WRITE_SET_WE_LOW_E2;
		WRITE_SET_WE_LOW_E2:	nstate = WRITE_SET_WE_HIGH_E2;
		WRITE_SET_WE_HIGH_E2:	nstate = WRITE_SET_CE_HIGH_E2;
		WRITE_SET_CE_HIGH_E2:	nstate = WRITE_SET_ADD_CE_E3;
// erase 3rd
		WRITE_SET_ADD_CE_E3:	nstate = WRITE_SET_WE_LOW_E3;
		WRITE_SET_WE_LOW_E3:	nstate = WRITE_SET_WE_HIGH_E3;
		WRITE_SET_WE_HIGH_E3:	nstate = WRITE_SET_CE_HIGH_E3;
		WRITE_SET_CE_HIGH_E3:	nstate = WRITE_SET_ADD_CE_E4;
// erase 4th
		WRITE_SET_ADD_CE_E4:	nstate = WRITE_SET_WE_LOW_E4;
		WRITE_SET_WE_LOW_E4:	nstate = WRITE_SET_WE_HIGH_E4;
		WRITE_SET_WE_HIGH_E4:	nstate = WRITE_SET_CE_HIGH_E4;
		WRITE_SET_CE_HIGH_E4:	nstate = WRITE_SET_ADD_CE_E5;
// erase 5th
		WRITE_SET_ADD_CE_E5:	nstate = WRITE_SET_WE_LOW_E5;
		WRITE_SET_WE_LOW_E5:	nstate = WRITE_SET_WE_HIGH_E5;
		WRITE_SET_WE_HIGH_E5:	nstate = WRITE_SET_CE_HIGH_E5;
		WRITE_SET_CE_HIGH_E5:	nstate = WRITE_SET_ADD_CE_E6;
// erase 6th
		WRITE_SET_ADD_CE_E6:	nstate = WRITE_SET_WE_LOW_E6;
		WRITE_SET_WE_LOW_E6:	nstate = WRITE_SET_WE_HIGH_E6;
		WRITE_SET_WE_HIGH_E6:	nstate = WRITE_SET_CE_HIGH_E6;
		WRITE_SET_CE_HIGH_E6:	nstate = DONE;
// reset
		WRITE_SET_ADD_CE_R:		nstate = WRITE_SET_WE_LOW_R;
		WRITE_SET_WE_LOW_R:		nstate = WRITE_SET_WE_HIGH_R;
		WRITE_SET_WE_HIGH_R:	nstate = WRITE_SET_CE_HIGH_R;
		WRITE_SET_CE_HIGH_R:	nstate = DONE;
//
		DONE:
			if (enable) nstate = DONE;
			else nstate = WAIT;
		default:	nstate = WAIT;
	endcase
//
reg[20:0] flash_addr_save;
always @ (posedge clock or negedge resetn)
	if (~resetn) begin
		inprog <= 0;
		flash_addr <= 0;
		flash_addr_save <= 0;
		ack <= 0;
		flash_rw <= 1;
		FlashDataWrite <= 0;
		vme_data_out <= 0;
		rce <= 1;
		rwe <= 1;
		roe <= 1;
		end
	else case (state)
		WAIT: begin
			rce <= #1 1;
			rwe <= #1 1;
			roe <= #1 1;
			inprog <= #1 0;
			FlashDataWrite <= #1 0;
			vme_data_out <= #1 0;
			ack <= #1 0;
			flash_rw <= #1 vme_write;
			end
		ENABLE: begin
			inprog <= #1 ~reset_address;
			end
		READ_INTERNAL_ADDRESS:
			vme_data_out <= #1 ({3'h0,flash_addr_save});
		RESET:			flash_addr_save <= #1 0;
		INCREMENT:		flash_addr_save <= #1 flash_addr_save + 1;
// read
		READ_SET_ADD:	flash_addr <= #1 flash_addr_save;
		READ_SET_CE:	rce <= #1 0;
		READ_SET_OE:	roe <= #1 0;
		READ_LATCH:		vme_data_out <= #1 ({16'h0, FlashDataRead});
// program 1st
		WRITE_SET_ADD_CE1: begin
			flash_addr <= #1 20'h555;
			FlashDataWrite[7:0] <= #1 8'haa;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW1:	rwe <= #1 0;
		WRITE_SET_WE_HIGH1:	rwe <= #1 1;
		WRITE_SET_CE_HIGH1:	rce <= #1 1;
// program 2nd
		WRITE_SET_ADD_CE2: begin
			flash_addr <= #1 20'h2aa;
			FlashDataWrite[7:0] <= #1 8'h55;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW2:	rwe <= #1 0;
		WRITE_SET_WE_HIGH2:	rwe <= #1 1;
		WRITE_SET_CE_HIGH2:	rce <= #1 1;
// program 3rd
		WRITE_SET_ADD_CE3: begin
			flash_addr <= #1 20'h555;
			FlashDataWrite[7:0] <= #1 8'ha0;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW3:	rwe <= #1 0;
		WRITE_SET_WE_HIGH3:	rwe <= #1 1;
		WRITE_SET_CE_HIGH3:	rce <= #1 1;
// program 4th
		WRITE_SET_ADD_CE4: begin
			flash_addr <= flash_addr_save;
			FlashDataWrite[7:0] <= #1 LocData[7:0];
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW4:	rwe <= #1 0;
		WRITE_SET_WE_HIGH4:	rwe <= #1 1;
		WRITE_SET_CE_HIGH4:	rce <= #1 1;
// erase 1st
		WRITE_SET_ADD_CE_E1: begin
			flash_addr <= #1 20'h555;
			FlashDataWrite[7:0] <= #1 8'haa;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW_E1:	rwe <= #1 0;
		WRITE_SET_WE_HIGH_E1:	rwe <= #1 1;
		WRITE_SET_CE_HIGH_E1:	rce <= #1 1;
// erase 2nd
		WRITE_SET_ADD_CE_E2: begin
			flash_addr <= #1 20'h2aa;
			FlashDataWrite[7:0] <= #1 8'h55;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW_E2:	rwe <= #1 0;
		WRITE_SET_WE_HIGH_E2:	rwe <= #1 1;
		WRITE_SET_CE_HIGH_E2:	rce <= #1 1;
// erase 3rd
		WRITE_SET_ADD_CE_E3: begin
			flash_addr <= #1 20'h555;
			FlashDataWrite[7:0] <= #1 8'h80;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW_E3:	rwe <= #1 0;
		WRITE_SET_WE_HIGH_E3:	rwe <= #1 1;
		WRITE_SET_CE_HIGH_E3:	rce <= #1 1;
// erase 4th
		WRITE_SET_ADD_CE_E4: begin
			flash_addr <= #1 20'h555;
			FlashDataWrite[7:0] <= #1 8'haa;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW_E4:	rwe <= #1 0;
		WRITE_SET_WE_HIGH_E4:	rwe <= #1 1;
		WRITE_SET_CE_HIGH_E4:	rce <= #1 1;
// erase 5th
		WRITE_SET_ADD_CE_E5: begin
			flash_addr <= #1 20'h2aa;
			FlashDataWrite[7:0] <= #1 8'h55;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW_E5:	rwe <= #1 0;
		WRITE_SET_WE_HIGH_E5:	rwe <= #1 1;
		WRITE_SET_CE_HIGH_E5:	rce <= #1 1;
// erase 6th
		WRITE_SET_ADD_CE_E6: begin
			flash_addr <= #1 20'h555;
			FlashDataWrite[7:0] <= #1 8'h10;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW_E6:	rwe <= #1 0;
		WRITE_SET_WE_HIGH_E6:	rwe <= #1 1;
		WRITE_SET_CE_HIGH_E6:	rce <= #1 1;
// reset
		WRITE_SET_ADD_CE_R: begin
			flash_addr <= #1 20'h0;
			FlashDataWrite[7:0] <= #1 8'hf0;
			rce <= #1 0;
			end
		WRITE_SET_WE_LOW_R:	rwe <= #1 0;
		WRITE_SET_WE_HIGH_R:	rwe <= #1 1;
		WRITE_SET_CE_HIGH_R:	rce <= #1 1;
//
		DONE: begin
			rce <= #1 1;
			rwe <= #1 1;
			roe <= #1 1;
			ack <= #1 1;
			end
	endcase


endmodule
