`timescale 1ns / 1ps
`include "define.v"
////////////////////////////////////////////////////////////////////////////////
/* This chip is a bridge between the Local Bus and the Flash devices 
 on the mezzanine.  
 We re-use stuff and names from the HTR Altera v1E project, other
 than the HTR Altera bridges between VMEbus to LocalBus, while
 this chip bridges between LocalBus to Flash-lines.
 The chip is mostly a "Flash handler" and a Xilix-configuator.
 On the demo board there are two FPGAs and two Flash devices, but
 on the final board, there is only one FPGA and one Flash device.
 We only want to demonstrate functionality for the final board,
 so the 2nd FPGA and Flash will be configured by cable only.
 We assume that this chip receive one LocalBus Chip Select signal,
 that is the chip implements one LocalBus slave.
 So far we assume that when a Chip Select is active, we "DO_FLASH",
 so no other enable signal is needed.
*/

/* Logic occupancy: 35% FlashHandler + 8% load2xi + 25% XilinxConfig ~ 70%
   Logic occupancy can be reduced eliminating crc, useless reset signals 
   and reset conditions.
   With disconnected crc, logic occupancy is 52%  */

module HFTbridge(
 input clk40_P, clk40_N,  // differential CLK input
 input BC0_p, BC0_n,     //differential BC0
 input TTC_CK40Des1, 	TTC_BCntRes,
 input [7:2] TTC_Brcst,
 input TTC_BrcstStr, // BrcstStr1, related to CK40Des1 and Brcst[5:2]
 //   Local Bus  port:
 input LocCS_b, 
 input LocWRITE_b,  // LOW = Write , HIGH = Read
 input [`LB_Addr_Size-1:0] LocAddr,  
 inout [`LB_Data_Size-1:0] LocData,  // if future could be inout (crc, etc)
 //   Connection to the  Flash memory device:
 output Flash1CE_b, Flash1OE_b, Flash1WE_b,
 inout  [7:0] FlashData, 
 output [20:0] FlashAddr,
 // Connections to the two Xilinx FPGA:
 input  ConfigDone1,  ConfigInit1, 
 output ConfigProg1,  ConfigClk,    // Hi-Z when using cable.
 output ConfigCS1,    ConfigWr_b                );

 wire do1,  FlashAck;
 wire [7:0] FlashDataWrite;
 wire [20:0] FlashAddrHandler, FlashAddrConfig;
 reg  [3:0]  ClkDiv;

 always @(posedge TTC_CK40Des1) ClkDiv <= ClkDiv + 1;

FlashHandler     FlashHandler (
    .clock(ClkDiv[2]),               .resetn(1'b1), 
    .enable(~LocCS_b),               .LocCS(~LocCS_b),         
    .LocData(LocData[7:0]),          .LocAddr(LocAddr), 
    .FlashDataRead(FlashData), 
    .vme_write(~LocWRITE_b),         .X_CONFIG_CS1(ConfigCS1), 
    .ack(FlashAck),   // there is no acknowledge to LocalBus Master.
    .flash_rw(flash_rw),           .inprog(FlashInprog), 
    .FLASH_1_CE(Flash1CE_b), 
    .FLASH_1_OE(Flash1OE_b), 
    .FLASH_1_WE(Flash1WE_b), 
    .FlashDataWrite(FlashDataWrite),       .flash_addr(FlashAddrHandler), 
    .vme_data_out(),   //address read-back problem, due to lack of LocData lines.
    .state()        );

assign FlashData = (FlashInprog & ~flash_rw) ? FlashDataWrite : 8'bZ;

/*  On XilinxConfig: "enable" replaces "load" and "config_vme".
 so the module is enabled only at powerup and after writing into the Flash */
XilinxConfig XilinxConfig1 (   
    .clock(ClkDiv[3]),               .resetn(1'b1), 
    .enable(do1 | FlashAck),        .data(LocData[7:0]), 
    .done(ConfigDone1),            .init(ConfigInit1), 
               //  outputs: 
    .addr(FlashAddrConfig),        .prog_b(prog_b),             
    .cs_b(ConfigCS1),              .inprog(ConfigInprog),             
    .cdone(done1),                 .id()    );

load2xi load2xi (    
    .clock(TTC_CK40Des1), 
    .enable1(1'b1),        .enable2(1'b0),   //  2nd Flash and Xilinx disabled
    .done1(done1),         .done2(1'b0), 
    .do1(do1),             .do2()         ); 

assign FlashAddr = (~LocCS_b) ?  FlashAddrHandler :  FlashAddrConfig;

assign ConfigProg1 = (ConfigInprog & do1) ? prog_b : 1'bZ;
assign ConfigClk   = ConfigInprog  ? ClkDiv[3] : 1'bZ;
assign ConfigWr_b  = 1'b0;   // always write-mode, as we do not read-back.

endmodule