`timescale 1ns / 1ps

// this module handles configuration of the xilinx(s) parts from the flash
// the module programs the xilinx using SelectMAP mode for parallel input,
// the input coming from the flash over the localbus data lines. the flash
// address bits are also localbus address lines.  these lines change on the
// negedge of the config clock, whereas the posedge causes the xilinx part
// to register the new data.

module XilinxConfig(
 input  clock, resetn, enable, done, init,
 input  [7:0] data,
 output prog_b, cs_b,  
 output reg [20:0] addr,
 output reg inprog, cdone,	
 output [23:0] id               );

// invert signals so they start up correctly
reg prog, cs;
assign prog_b = ~prog;
assign cs_b = ~cs;

// two types of reset
wire my_reset = resetn;
wire clocking_done;

// make a state machine
reg[3:0] state, nstate;
parameter[3:0] WAIT=0,  CS_SET=1, PROG=2,  PROG1=3,      PROG2=4, 
            	INIT=5,  WRITE=6,  DELAY=7, CLOCK_DATA=8, DONE=9, 
	            POST_CLOCK1=10,    POST_CLOCK2=11;
always @ (posedge clock)
	if (~my_reset) state <= WAIT;
	else state <= #1 nstate;
assign clocking_done = (addr == 21'b111111111111111111111);

always @ (enable or init or done or clocking_done or state)
	case (state)
		WAIT:		if (enable) nstate = CS_SET;
		CS_SET:  nstate = PROG;
		PROG:	   nstate = PROG1;
		PROG1:	nstate = PROG2;
		PROG2:	nstate = INIT;
		INIT:    if (init) nstate = WRITE;
		WRITE:	nstate = DELAY;
		DELAY:   nstate = CLOCK_DATA;

// these chips will hold 21 bits, but need only about half that.
// so when done comes back, that's enough
		CLOCK_DATA: 
			if (clocking_done) nstate = DONE;
			else nstate = CLOCK_DATA;
//
// wait here until "done" comes back
//
		DONE: 
			if (done) nstate = POST_CLOCK1;
			else nstate = DONE;
		POST_CLOCK1:
			if (enable) nstate = POST_CLOCK1;
			else nstate = WAIT;
		default: nstate = WAIT;
	endcase
//
// id is the "checksum", or "crc", or "parity", depending on
// what you put into the code.  basically, this is a single
// number calculated from all of the bytes that go into the
// flash, and serves as a verification that the xilinx FPGA
// firmware version indeed is unique.  it helps us in the case
// where you change the firmware but not the version!
//
reg load_crc;
reg reset_crc;
always @ (posedge clock)
	if (~my_reset) begin
		load_crc <= 0;
		reset_crc <= 0;
		end
	else case (state)
		WAIT: begin
			reset_crc <= #1 0;
			load_crc <= #1 0;
			end
		CS_SET: reset_crc <= #1 1;
		PROG:	reset_crc <= #1 0;
		DELAY:	load_crc <= #1 1;
		DONE:	load_crc <= #1 0;
		endcase
crc24 CRC (.CLK(clock),.RESET(reset_crc),.LOAD(load_crc),.DATAIN(data),.CRCOUT(id));
//
// inprog bit enables tristate outputs
//
always @ (negedge clock)
	if (~my_reset) inprog <= 0;
	else case (state)
		WAIT: inprog <= 0;
		PROG: inprog <= #1 1;
		DONE: inprog <= #1 0;
		default:  inprog <= inprog;
	endcase
//
// make 21-bit address counter
//
always @ (negedge clock)
	if (~my_reset) addr <= 0;
	else case (state)
		WAIT: addr <= 0;
		CLOCK_DATA: addr <= #1 addr + 1;
		default: addr <= addr;
	endcase

// here are the rest of the signals:  cs and prog
always @ (negedge clock)
   case (state)
		WAIT: begin
			prog <= #1 0;    cs <= #1 0;  cdone <= #1 0;
			end
		CS_SET: cs <= #1 1;  
		PROG:	  prog <= #1 1;
		PROG2:  prog <= #1 0;
		DELAY: begin		end
		DONE: begin
			cs <= #1 0;
			prog <= #1 0;
			end
		POST_CLOCK1: cdone <= #1 1;
	endcase

endmodule
